#ifndef SUBITEM_VALUE_PER_ITEM_HPP
#define SUBITEM_VALUE_PER_ITEM_HPP

#include <Kokkos_StaticCrsGraph.hpp>
#include <ItemType.hpp>
#include <PastisAssert.hpp>

#include <ConnectivityMatrix.hpp>
#include <IConnectivity.hpp>

template <typename DataType,
          ItemType sub_item_type,
          ItemType item_type,
          typename Allowed=void>
class SubItemValuePerItem;

template <typename DataType,
          ItemType sub_item_type,
          ItemType item_type>
class SubItemValuePerItem<DataType,
                          sub_item_type,
                          item_type,
                          std::enable_if_t<sub_item_type != item_type>>
{
 public:
  static const ItemType item_t{item_type};
  static const ItemType sub_item_t{sub_item_type};
 private:
  bool m_is_built{false};

  ConnectivityMatrix::HostRowType m_host_row_map;
  Kokkos::View<DataType*> m_values;

  // Allows const version to access our data
  friend SubItemValuePerItem<std::add_const_t<DataType>,
                             sub_item_type,
                             item_type>;

 public:
  class SubView
  {
   private:
    KOKKOS_RESTRICT DataType* const m_sub_values;
    const size_t m_size;
   public:
    KOKKOS_INLINE_FUNCTION
    const DataType& operator[](const size_t& i) const
    {
      Assert(i<m_size);
      return m_sub_values[i];
    }

    KOKKOS_FORCEINLINE_FUNCTION
    DataType& operator[](const size_t& i)
    {
      Assert(i<m_size);
      return m_sub_values[i];
    }

    KOKKOS_INLINE_FUNCTION
    const size_t& size() const
    {
      return m_size;
    }

    SubView(const SubView&) = delete;

    KOKKOS_INLINE_FUNCTION
    SubView(SubView&&) = default;

    KOKKOS_INLINE_FUNCTION
    SubView(const Kokkos::View<DataType*>& values,
            const size_t& begin,
            const size_t& end)
        : m_sub_values(&(values[begin])),
          m_size(end-begin)
    {
      Assert(begin<=end);
      Assert(end<=values.extent(0));
    }
  };

  KOKKOS_FORCEINLINE_FUNCTION
  const bool& isBuilt() const
  {
    return m_is_built;
  }

  KOKKOS_FORCEINLINE_FUNCTION
  DataType& operator()(const size_t& j, const size_t& r)
  {
    Assert(m_is_built);
    return m_values[m_host_row_map(j)+r];
  }

  // Following Kokkos logic, these classes are view and const view does allow
  // changes in data
  KOKKOS_FORCEINLINE_FUNCTION
  DataType& operator()(const size_t& j, const size_t& r) const
  {
    Assert(m_is_built);
    return m_values[m_host_row_map(j)+r];
  }

  KOKKOS_INLINE_FUNCTION
  size_t numberOfValues() const
  {
    Assert(m_is_built);
    return m_values.extent(0);
  }

  KOKKOS_FORCEINLINE_FUNCTION
  DataType& operator[](const size_t& i)
  {
    Assert(m_is_built);
    return m_values[i];
  }

  // Following Kokkos logic, these classes are view and const view does allow
  // changes in data
  KOKKOS_FORCEINLINE_FUNCTION
  DataType& operator[](const size_t & i) const
  {
    Assert(m_is_built);
    return m_values[i];
  }

  KOKKOS_INLINE_FUNCTION
  size_t numberOfItems() const
  {
    Assert(m_is_built);
    Assert(m_host_row_map.extent(0) != 0);
    return m_host_row_map.extent(0);
  }

  KOKKOS_INLINE_FUNCTION
  size_t numberOfSubValues(const size_t& i_cell) const
  {
    Assert(m_is_built);
    return m_host_row_map(i_cell+1)-m_host_row_map(i_cell);
  }

  KOKKOS_INLINE_FUNCTION
  SubView itemValues(const size_t& i_cell)
  {
    Assert(m_is_built);
    const auto& cell_begin = m_host_row_map(i_cell);
    const auto& cell_end = m_host_row_map(i_cell+1);
    return SubView(m_values, cell_begin, cell_end);
  }

  // Following Kokkos logic, these classes are view and const view does allow
  // changes in data
  KOKKOS_INLINE_FUNCTION
  SubView itemValues(const size_t& i_cell) const
  {
    Assert(m_is_built);
    const auto& cell_begin = m_host_row_map(i_cell);
    const auto& cell_end = m_host_row_map(i_cell+1);
    return SubView(m_values, cell_begin, cell_end);
  }

  template <typename DataType2>
  KOKKOS_INLINE_FUNCTION
  SubItemValuePerItem&
  operator=(const SubItemValuePerItem<DataType2, sub_item_type, item_type>& sub_item_value_per_item)
  {
    // ensures that DataType is the same as source DataType2
    static_assert(std::is_same<std::remove_const_t<DataType>, std::remove_const_t<DataType2>>(),
                  "Cannot assign SubItemValuePerItem of different type");
    // ensures that const is not lost through copy
    static_assert(((std::is_const<DataType2>() and std::is_const<DataType>())
                   or not std::is_const<DataType2>()),
                  "Cannot assign const  SubItemValuePerItem to a non const SubItemValuePerItem");
    m_host_row_map = sub_item_value_per_item.m_host_row_map;
    m_values = sub_item_value_per_item.m_values;

    m_is_built = sub_item_value_per_item.m_is_built;

    return *this;
  }

  template <typename DataType2>
  KOKKOS_INLINE_FUNCTION
  SubItemValuePerItem(const SubItemValuePerItem<DataType2, sub_item_type, item_type>& sub_item_value_per_item)
  {
    this->operator=(sub_item_value_per_item);
  }

  SubItemValuePerItem() = default;

  SubItemValuePerItem(const IConnectivity& connectivity)
      : m_is_built{true}
  {
    ConnectivityMatrix connectivity_matrix
        = connectivity.getMatrix(item_type, sub_item_type);

    m_host_row_map = connectivity_matrix.rowsMap();
    m_values = Kokkos::View<DataType*>("values", connectivity_matrix.numEntries());
  }

  ~SubItemValuePerItem() = default;
};

// Item values at nodes

template <typename DataType>
using NodeValuePerEdge = SubItemValuePerItem<DataType, ItemType::node, ItemType::edge>;

template <typename DataType>
using NodeValuePerFace = SubItemValuePerItem<DataType, ItemType::node, ItemType::face>;

template <typename DataType>
using NodeValuePerCell = SubItemValuePerItem<DataType, ItemType::node, ItemType::cell>;

// Item values at edges

template <typename DataType>
using EdgeValuePerNode = SubItemValuePerItem<DataType, ItemType::edge, ItemType::node>;

template <typename DataType>
using EdgeValuePerFace = SubItemValuePerItem<DataType, ItemType::edge, ItemType::face>;

template <typename DataType>
using EdgeValuePerCell = SubItemValuePerItem<DataType, ItemType::edge, ItemType::cell>;

// Item values at faces

template <typename DataType>
using FaceValuePerNode = SubItemValuePerItem<DataType, ItemType::face, ItemType::node>;

template <typename DataType>
using FaceValuePerEdge = SubItemValuePerItem<DataType, ItemType::face, ItemType::edge>;

template <typename DataType>
using FaceValuePerCell = SubItemValuePerItem<DataType, ItemType::face, ItemType::cell>;

// Item values at cells

template <typename DataType>
using CellValuePerNode = SubItemValuePerItem<DataType, ItemType::cell, ItemType::node>;

template <typename DataType>
using CellValuePerEdge = SubItemValuePerItem<DataType, ItemType::cell, ItemType::edge>;

template <typename DataType>
using CellValuePerFace = SubItemValuePerItem<DataType, ItemType::cell, ItemType::face>;

#endif // SUBITEM_VALUE_PER_ITEM_HPP
